/*============================================================================
 *
 *  Program:      CheckCRC.c
 *
 *  Version:      1.1 (15.12.94)
 *
 *  Purpose:      calculates a 32-bit CRC checksum of a file and compares
 *                it with a checksum given in the file's first line
 *
 *  Parameters:   CheckCRC <listfilename>
 *
 *  Returns:      0 : The CRC-sum is correct
 *                5 : The file has no CRC-sum
 *                10: The file already has a CRC-sum, which is wrong
 *                20: Syntax-, I/O- or Memory-Error.
 *
 *  Comment:      The CRC-algorithm was specified by W. D. Schwaderer in 
 *						his book "C Programmer's Guide to NetBIOS" 
 *						Howard W. Sams & Company, First Edition 1988                                                 
 *
 *  Author:       Andre Bernhardt <ab@imdb.com>
 *
 *  Copyright:    (w) Andre Bernhardt 1994 - 1996
 *                (c) Internet MovieDatabase Limited 1996
 *
 *       This file is part of the Internet MovieDatabase project.
 *
 *  The  MovieDatabase  FAQ contains more information on the whole project.
 *  For   a   copy   send  an  e-mail   with  the  subject  "HELP  FAQ"  to
 *  <movie@ibmpcug.co.uk>.
 *
 *  Permission  is  granted  to make and distribute verbatim copies of this
 *  package  provided  the  copyright notice and this permission notice are
 *  preserved  on  all  copies  and the package is distributed in unaltered
 *  archive  form only.  It is not allowed to modify the source code and/or
 *  redistribute  modified  copies  of  it  and/or  the executables without
 *  written permission of the author.
 *
 *  If  you need to make a change to the source-code in order to be able to
 *  use the package, you have to notify the author.
 *
 *  No guarantee of any kind is given that the programs and scripts in this
 *  package  are  100%  reliable.  You are using this material at your  own
 *  risk.   The  author  cannot be made responsible for any damage which is
 *  caused by using these programs.
 *
 *  This  package  is  freely  distributable,  but still copyright by Andre
 *  Bernhardt.
 *
 *  None  of  the programs or scripts nor the source code (nor parts of it)
 *  may  be  included  or  used  in  commercial  programs unless by written
 *  permission from the author.
 *
 *============================================================================
 */

#include "IMDB.h"

/* Return values */
#define RET_CRC_OK         0
#define RET_NO_CRC         5
#define RET_CRC_ERROR      10
#define RET_ERROR          20

/* integer bit structure for initialize CRC-32 table */
/* warning: it is important that <i1> is the lowest bit in the structure */
typedef struct
 {
#ifndef JTL
   ULONG i1:1,i2:1,i3:1,i4:1,i5:1,i6:1,i7:1,i8:1;
#else
   UBYTE i8:1,i7:1,i6:1,i5:1,i4:1,i3:1,i2:1,i1:1;
#endif
 }
IBIT;

/* long bit structure for initialize CRC-32 table */
/* warning: it is important that <b1> is the lowest bit in the structure */
typedef struct
 {
#ifndef JTL
   ULONG  b1:1, b2:1, b3:1, b4:1, b5:1, b6:1, b7:1, b8:1;
   ULONG  b9:1,b10:1,b11:1,b12:1,b13:1,b14:1,b15:1,b16:1;
   ULONG b17:1,b18:1,b19:1,b20:1,b21:1,b22:1,b23:1,b24:1;
   ULONG b25:1,b26:1,b27:1,b28:1,b29:1,b30:1,b31:1,b32:1;
#else
   ULONG b32:1,b31:1,b30:1,b29:1,b28:1,b27:1,b26:1,b25:1;
   ULONG b24:1,b23:1,b22:1,b21:1,b20:1,b19:1,b18:1,b17:1;
   ULONG b16:1,b15:1,b14:1,b13:1,b12:1,b11:1,b10:1, b9:1;
   ULONG  b8:1, b7:1, b6:1, b5:1, b4:1, b3:1, b2:1, b1:1;
#endif
 }
LBIT;

#define VERSION "CheckCRC 1.1 (15.03.96)"
static const char version[] ="$VER: "VERSION;

/*-----------------------------------------------------------------------------
 * Procedure:   init_crc
 *
 * Purpose:     This function initializes the CRC-32 algorithm by setting a private
 *              pre  calculation  table  in the memory.  This table is used
 *              for all further CRC-32 calculations.
 *
 * Parameters:  none
 *
 * Returns:     Pointer to CRC-Table
 *-----------------------------------------------------------------------------
 */

ULONG *init_crc(void)
 {
  ULONG *pCrcTab = NULL;
  int i;
  IBIT ib;
  LBIT lb;
  
  /* allocate memory for table */
  if (pCrcTab = malloc(256 * sizeof(ULONG)))
   {  
    /* initialize table with calculated bits */
    for (i=0; i<256; i++)
     {
      /* calculate each bit */
      *(UBYTE*)&ib=i;
      *(ULONG*)&lb=0; /* all bits sb.b_ are zeroed out now */
    
      /* calculate bits in <lb> from bits in <ib> */
      lb.b32=ib.i2^ib.i8;
      lb.b31=ib.i1^ib.i2^ib.i7^ib.i8;
      lb.b30=ib.i1^ib.i2^ib.i6^ib.i7^ib.i8;
      lb.b29=ib.i1^ib.i5^ib.i6^ib.i7;
      lb.b28=ib.i2^ib.i4^ib.i5^ib.i6^ib.i8;
      lb.b27=ib.i1^ib.i2^ib.i3^ib.i4^ib.i5^ib.i7^ib.i8;
      lb.b26=ib.i1^ib.i2^ib.i3^ib.i4^ib.i6^ib.i7;
      lb.b25=ib.i1^ib.i3^ib.i5^ib.i6^ib.i8;
      lb.b24=ib.i4^ib.i5^ib.i7^ib.i8;
      lb.b23=ib.i3^ib.i4^ib.i6^ib.i7;
      lb.b22=ib.i3^ib.i5^ib.i6^ib.i8;
      lb.b21=ib.i4^ib.i5^ib.i7^ib.i8;
      lb.b20=ib.i2^ib.i3^ib.i4^ib.i6^ib.i7^ib.i8;
      lb.b19=ib.i1^ib.i2^ib.i3^ib.i5^ib.i6^ib.i7;
      lb.b18=ib.i1^ib.i2^ib.i4^ib.i5^ib.i6;
      lb.b17=ib.i1^ib.i3^ib.i4^ib.i5;
      lb.b16=ib.i3^ib.i4^ib.i8;
      lb.b15=ib.i2^ib.i3^ib.i7;
      lb.b14=ib.i1^ib.i2^ib.i6;
      lb.b13=ib.i1^ib.i5;
      lb.b12=ib.i4;
      lb.b11=ib.i3;
      lb.b10=ib.i8;
      lb.b9 =ib.i2^ib.i7^ib.i8;
      lb.b8 =ib.i1^ib.i6^ib.i7;
      lb.b7 =ib.i5^ib.i6;
      lb.b6 =ib.i2^ib.i4^ib.i5^ib.i8;
      lb.b5 =ib.i1^ib.i3^ib.i4^ib.i7;
      lb.b4 =ib.i2^ib.i3^ib.i6;
      lb.b3 =ib.i1^ib.i2^ib.i5;
      lb.b2 =ib.i1^ib.i4;
      lb.b1 =ib.i3;
      
      /* store calculated ULONG value */
      pCrcTab[i]=*(ULONG*)&lb;
     }
   }
  return pCrcTab;
 }

/*-----------------------------------------------------------------------------
 * Procedure:   main
 *
 * Parameters:  nb_args, filename
 *
 * Returns:     Message, if CRC-Checksum is Okay
 * 
 * Comments:    Before the CRC-32 for a complete sequence is calculated, each
 *              bit  in  the  CRC  value must be set to 1 (set value to -1)
 *              After   calculation   the   complete   sequence,  the  real
 *              CRC-sum-value  is  the  1-complete  of  the  calculated CRC
 *              value.  Use the "~"-operator to calculate.
 *-----------------------------------------------------------------------------
 */

int main(int argc, char *argv[])
{
 static char old_crc[16];
 static char new_crc[16];
 ULONG nCrc = 0xFFFFFFFFL;
 ULONG *pCrcTab;
 FILE *file;
 char *fname;
 int i, c, nIndex;

#ifdef SYS_AMIGA
  printf (VERSION" written by Andr Bernhardt;  1996 IMDb Ltd.\n"); 
#else
  printf (VERSION" written by Andr Bernhardt; (c) 1996 IMDb Ltd.\n"); 
#endif

 if (argc != 2)
  {
#ifdef SYS_AMIGA
   fprintf(stderr, "Template: CheckCRC LISTFILE/A\n");
#else
   fprintf(stderr, "Template: CheckCRC <listfile>\n");
#endif
   exit (RET_ERROR);
  }

 fname = argv[1];

 if (file = fopen(fname,"r"))
  {
   /* Read the first 15 bytes (eg.: "CRC: 0x9FED5FE2") and skip line*/
   i = 0;
   while ((i < 15) && (EOF != (c = fgetc (file))) && ('\n' != c))
    old_crc [i++] = c;
   old_crc [i] = '\0';

   /* skip line */
   while ((EOF != (c = fgetc (file))) && ('\n' != c));

   /* check for "CRC: " */     
   if (0 == strncmp (old_crc, "CRC: ", strlen("CRC: "))) 
    {
     if (pCrcTab = init_crc())
      {
       /* Create the CRC-checksum for the current file */
       /* calculate until EOF */
       while (EOF != (c = fgetc (file)))
        {
         /* calculate CRC for next byte */
         nIndex = (int) ((nCrc ^ c) & 0x000000FFL);
         nCrc = ((nCrc >> 8) & 0x00FFFFFFL) ^ pCrcTab[nIndex];
        }
      }
     else
      {
       fprintf(stderr, "Failed to create CRC-Tab\n");
       fclose (file);
       exit (RET_ERROR);
      }
     fclose (file);

     /* Free CRC-Tab */
     free(pCrcTab);

     /* compare CRC */
     sprintf(new_crc, "CRC: 0x%08X", nCrc);
     if (0 == strcmp (old_crc, new_crc))
      {
       fprintf (stderr, "CRC-Checksum O.K.\n");
       exit (RET_CRC_OK);
      }
     else
      {
       fprintf (stderr, "CRC-Checksum Error in file %s\n", fname);
       exit (RET_CRC_ERROR);
      }
    }
   else
    {
     fprintf(stderr, "File does not contain CRC-Checksum.\n");
     fclose (file);
     exit (RET_NO_CRC);
    }
  }
 else
  {
   fprintf(stderr, "Could not open %s\n",fname);
   exit (RET_ERROR);
  }
}
